#lang sicp

(define (deriv exp var)
  (cond ((number? exp) 0)
        ((variable? exp)
         (if (same-variable? exp var) 1 0))
        ((sum? exp)
         (make-sum (deriv (addend exp) var)
                   (deriv (augend exp) var)))
        ((product? exp)
         (make-sum
           (make-product (multiplier exp)
                         (deriv (multiplicand exp) var))
           (make-product (deriv (multiplier exp) var)
                         (multiplicand exp))))
        ((exponentiation? exp) ;;(make-product (exponent exp) (deriv (base exp) var))
         (let ((e (exponent exp))
               (b (base exp)))
           (make-product e
                         (make-product
                          (make-exponentiation b (- e 1))
                          (deriv b var)))))
        (else
         (error "unknown expression type -- DERIV" exp))))

;; representing algebraic expressions

(define (variable? x) (symbol? x))

(define (same-variable? v1 v2)
  (and (variable? v1) (variable? v2) (eq? v1 v2)))

;;(define (make-sum a1 a2) (list '+ a1 a2))

;;(define (make-product m1 m2) (list '* m1 m2))

(define (more-than-2-terms x) (pair? (cdddr x)))

;;(define (sum? x)
;;  (and (pair? x) (eq? (car x) '+)))
;;
;;(define (addend s) (cadr s))
;;
;;(define (augend s)
;;  (if (more-than-2-terms s)
;;      (cons '+ (cddr s))
;;      (caddr s)))
;;
;;(define (product? x)
;;  (and (pair? x) (eq? (car x) '*)))
;;
;;(define (multiplier p) (cadr p))
;;
;;(define (multiplicand p)
;;  (if (more-than-2-terms p)
;;      (cons '* (cddr p))
;;      (caddr p)))


;: (deriv '(+ x 3) 'x)
;: (deriv '(* x y) 'x)
;: (deriv '(* (* x y) (+ x 3)) 'x)


;; With simplification

;;(define (make-sum a1 a2)
;;  (cond ((=number? a1 0) a2)
;;        ((=number? a2 0) a1)
;;        ((and (number? a1) (number? a2)) (+ a1 a2))
;;        (else (list '+ a1 a2))))

(define (=number? exp num)
  (and (number? exp) (= exp num)))

;;(define (make-product m1 m2)
;;  (cond ((or (=number? m1 0) (=number? m2 0)) 0)
;;        ((=number? m1 1) m2)
;;        ((=number? m2 1) m1)
;;        ((and (number? m1) (number? m2)) (* m1 m2))
;;        (else (list '* m1 m2))))


;; Exercise 2.56

(define (make-exponentiation b p) 
 (cond ((=number? p 0) 1)
       ((=number? p 1) b)
       ((and (number? b) (number? p)) (expt b p))
       (else (list '** b p))))

(define (base e) (cadr e))
(define (exponent e) (caddr e))

(define (exponentiation? x) (and (pair? x) (eq? (car x) '**)))


;; Exercise 2.58

(define (make-sum a1 a2)
  (cond ((=number? a1 0) a2)
        ((=number? a2 0) a1)
        ((and (number? a1) (number? a2)) (+ a1 a2))
        ((product? a2) (list a1 '+ (multiplier a2) '* (multiplicand a2)))
        (else (list a1 '+ a2))))

(define (make-product m1 m2)
  (cond ((or (=number? m1 0) (=number? m2 0)) 0)
        ((=number? m1 1) m2)
        ((=number? m2 1) m1)
        ((and (number? m1) (number? m2)) (* m1 m2))
        (else (list m1 '* m2))))

(define (sum? x)
  (and (pair? x) (eq? (cadr x) '+)))

(define (addend s) (car s))

;; Special case of multiplication distributing across addition.  It
;; should be possible to rewrite augend and/or multiplicand to deal
;; with this case

(define (augend s)
  (if (product? (cddr s))
      (cddr s)
      (caddr s)))

(define (product? x)
  (and (pair? x) (eq? (cadr x) '*)))

(define (multiplier p) (car p))

(define (multiplicand p) (caddr p))
